<?php declare(strict_types=1);
/**
 * Twig Extension Class: for handling the environment of Twig
 */

namespace Infomaker\Everyware\Twig;

use BadFunctionCallException;
use BadMethodCallException;
use RuntimeException;
use Twig_Loader_Chain;
use Twig_LoaderInterface;
use Twig_Function;
use Twig_SourceContextLoaderInterface;

/**
 * Class EveryTwig
 * @link    http://infomaker.se
 * @package Infomaker\Everyware\Twig
 */
class EveryTwig
{

    /**
     * Contains an instance of the parent for calls to certain functions
     *
     * @var Environment
     */
    private $twig;

    /**
     * Contains an instance of self
     *
     * @var EveryTwig
     */
   /*  private static $instance; */

    /**
     * Contains a list of registered loaders
     *
     * @var array
     */
    private $loaders = [];

    /**
     * Contains whether to use cache
     *
     * @var bool
     */
    private $cache = false;

    /**
     * Time in seconds to cache the output
     *
     * @var int
     */
    private $cacheTime = 60;

    /**
     * Add suffix so we don't have to delete it in deploy every time
     *
     * @var string
     */
    private $cacheKey = 'everyware_twig_compiled_cache';

    private static $instance;
    
    private function __construct()
    {
        if( defined('GIT_COMMIT') ) {
            $this->cacheKey .= '_' . GIT_COMMIT;
        }
    }

    /**
     * Create new instance if none exist
     *
     * @return ViewSetup
     */
    public static function getInstance(): EveryTwig
    {
        if (static::$instance === null) {
            static::$instance = new static;
        }

        return static::$instance;
    }

    /**
     *Initializes the EveryTwig-wrapper and Twig environment
     *
     * @param array $loaders
     * @param array $functions
     * @param array $filters
     * @param array $options
     *
     * @return Environment
     */
    public function init(
        $loader,
        array $functions = [],
        array $filters = [],
        array $options = ['debug' => false, 'cache' => false]
    ): Environment {

        // Handle options
        $this->handleOptions($options);

        $this->addInitialLoader($loader);
        
        $chain = new Twig_Loader_Chain($this->loaders);
        $this->twig = new Environment($chain, $options);

        $this->twig->addFunction($this->defaultPhpFunction());
        $this->twig->addExtension(new \Twig_Extension_Debug());

        foreach ($functions as $function) {
            $this->twig->addFunction($function);
        }
        foreach ($filters as $filter) {
            $this->twig->addFilter($filter);
        }

        return $this->twig;
    }

    /**
     * Handles the specific cache options and trims them away from the options array.
     *
     * @param array $options
     *
     * @return void
     */
    private function handleOptions(array &$options): void
    {
        if (isset($options['cachetime']) && (int)$options['cachetime'] > 0) {
            $this->cacheTime = (int)$options['cachetime'];
            unset($options['cachetime']);
        }

        if (isset($options['cache'])) {
            $this->cache = $options['cache'];
            // Add custom cache system to Twig
            $options['cache'] = $this->cache ? ObjectCacheSystem::setup($this->cacheKey, $this->cacheTime) : false;
        }
    }

    /**
     * Helper function to add a Twig_LoaderInterface loader on init
     *
     * @param Twig_LoaderInterface|Twig_SourceContextLoaderInterface $loader
     *
     * @return void
     */
    private function addInitialLoader(Twig_SourceContextLoaderInterface $loader): void
    {
        $this->loaders[] = $loader;
    }
    /**
     * Twig_SimpleFunction to be able to call default php functions from within twig templates
     *
     * @return Twig_Function
     */
    protected function defaultPhpFunction(): Twig_Function
    {
        $phpFunction = new Twig_Function('php_function', function ($name, $args = []) {
            if (!\function_exists($name)) {
                throw new BadFunctionCallException('Function dose not exist');
            }

            if (strpos($name, '::')) {
                $staticMethod = explode('::', $name);
                if (!method_exists($staticMethod[0], $staticMethod[1])) {
                    throw new BadMethodCallException('Method ' . $staticMethod[1] . ' can not be found on ' . $staticMethod[0]);
                }
            }

            return \call_user_func_array($name, $args);

        });

        return $phpFunction;
    }

    /**
     * Render function of EveryTwig, loads a specific template and works with cached compiled templates.
     *
     * @param string $template
     * @param array $context
     *
     * @return void
     * @throws \Twig_Error_Loader
     * @throws \Twig_Error_Runtime
     * @throws \Twig_Error_Syntax
     */
    public function render(string $template, array $context = null): void
    {
        $this->twig->display($template, $context ?? []);
    }

    /**
     * Return the view as rendered HTML by passed template
     *
     * @param string $template
     * @param array $context
     *
     * @return mixed
     * @throws \Twig_Error_Loader
     * @throws \Twig_Error_Runtime
     * @throws \Twig_Error_Syntax
     */
    public function view(string $template, array $context = null): string
    {
        return $this->twig->render($template, $context ?? []);
    }
}
