<?php declare(strict_types=1);

namespace Infomaker\Everyware\Twig;

use Twig_Error;
use Twig_Filter;
use Twig_Function;


class ViewSetup
{
    protected $cache = false;
    protected $cacheTime = 86400;
    protected $debug = true;

    /**
     * Contains functions that will be available in the templates.
     * If the same function name is added several times only the last one is kept.
     * @var array
     */
    protected $functions = [];

    /**
     * Contains filters that will be available in the templates.
     * If the same filters name is added several times only the last one is kept.
     * @var array
     */
    protected $filters = [];
    /**
     * Contains filters that will be available in the templates.
     * @var array
     */
    protected $globals = [];

    /**
     * Contains the FilesystemLoader aka all template-paths available in the templates.
     * @var FilesystemLoader
     */
    protected $loader;
    protected $everyTwig;

    private static $instance;

    private function __construct()
    {
        //Add global device variable
        if (\defined('EVERY_DEVICE')) {
            $this->addGlobal('device', EVERY_DEVICE);
        }
    }

      /**
     * Create new instance if none exist
     *
     * @return ViewSetup
     */
    public static function getInstance(): ViewSetup
    {
        if (static::$instance === null) {
            static::$instance = new static;
        }

        return static::$instance;
    }

    public function setLoader(FilesystemLoader $loader): void
    {
        $this->loader = $loader;
    }

    public function setEveryTwig(EveryTwig $everyTwig): void
    {
        $this->everyTwig = $everyTwig;
    }

    private function getLoader(): FilesystemLoader
    {
        return $this->loader;
    }

    public function getEveryTwig(): EveryTwig
    {
        return $this->everyTwig;
    }

    public function addFilter(string $name, $callable, array $options = null): self
    {
        $this->filters[$name] = new Twig_Filter($name, $callable, $options ?? []);

        return $this;
    }

    public function addFunction(string $name, $callable, array $options = null)
    {
        if( ! is_callable($callable, true)) {
            return false;
        }
        $this->functions[$name] = new Twig_Function($name, $callable, $options ?? []);

        return $this;
    }

    public function addGlobal(string $key, $value): self
    {
        $this->globals[$key] = $value;

        return $this;
    }

    public function addWpFunction(string $functionName, string $name = null): self
    {
        $name = ( ! empty($name) && $name !== null) ? $name : $functionName;
        $this->addFunction($name, $functionName);

        return $this;
    }

    public function init(array $options = []): void
    {
        $everyTwig = $this->getEveryTwig();
        $twig = $everyTwig->init($this->getLoader(), $this->functions, $this->filters,
            array_replace([
                'debug' => $this->debug,
                'cache' => $this->cache,
                'cachetime' => $this->cacheTime,
            ], $options));

        foreach ($this->globals as $key => $value) {
            $twig->addGlobal($key, $value);
        }
        //$everyTwig->addExtensions();
        $twig->setLexer(new \Twig_Lexer($twig));
        $twig->setParser(new \Twig_Parser($twig));
    }

    public function registerTwigFolder(string $namespace, string $folderPath, array $exclude = null): self
    {
        if (is_dir($folderPath)) {
            try {
                $this->getLoader()->addPathAndSubfolders($folderPath, $namespace, $exclude ?? []);
            } catch (Twig_Error $e) {
                new \Exception($e->getMessage());
            }
        }

        return $this;
    }
}
